import random
from typing import Optional

from aio_pika import Message, DeliveryMode
from aio_pika.abc import ExchangeType, AbstractChannel, AbstractQueue
from aiormq import DeliveryError, ChannelNotFoundEntity

from khnm.types import SenderT, QueueGetterT
from khnm.time import Clock, UtcClock
from khnm.types import SuccessT


def get_exchange_name(pipe_name: str) -> str:
    return f"khnm.ex.{pipe_name}"


def get_queue_name(pipe_name: str) -> str:
    return f"khnm.q.{pipe_name}"


async def declare_pipe(
    channel: AbstractChannel,
    name: str,
    size: Optional[int] = None,
    durable: bool = False,
) -> None:
    exchange_name = get_exchange_name(name)
    queue_name = get_queue_name(name)
    exchange = await channel.declare_exchange(exchange_name, ExchangeType.DIRECT)
    if size is not None:
        queue = await channel.declare_queue(
            name=queue_name,
            durable=durable,
            arguments={"x-max-length": size, "x-overflow": "reject-publish"},
        )
    else:
        queue = await channel.declare_queue(name=queue_name, durable=durable)
    await queue.bind(exchange, routing_key=queue_name)


async def send_message(
    channel: AbstractChannel, message: Message, pipe: str, persistent: bool = False
) -> SuccessT:
    if persistent:
        message.delivery_mode = DeliveryMode.PERSISTENT
    exchange = await channel.get_exchange(get_exchange_name(pipe), ensure=False)
    try:
        await exchange.publish(message, routing_key=get_queue_name(pipe))
    except DeliveryError as e:
        if e.frame.name == "Basic.Nack":
            return False
        raise
    return True


async def send_with_backoff(
    channel: AbstractChannel,
    sender: SenderT,
    message: Message,
    pipe: str,
    backoff_seconds: float = 0.1,
    max_retries: Optional[int] = None,
    exponential_backoff: bool = False,
    max_backoff_seconds: Optional[float] = None,
    apply_jitter: bool = False,
    persistent: bool = False,
    clock: Clock = UtcClock(),
) -> SuccessT:
    retries = 0
    sent = await sender(channel, message, pipe, persistent)
    while not sent and (max_retries is None or retries < max_retries):
        wait = _calculate_backoff(
            retries,
            backoff_seconds,
            exponential_backoff,
            max_backoff_seconds,
            apply_jitter,
        )
        await clock.sleep(wait)
        sent = await sender(channel, message, pipe, persistent)
        retries += 1
    return sent


async def get_queue(channel: AbstractChannel, queue: str) -> AbstractQueue:
    return await channel.get_queue(queue)


async def wait_for_pipe(
    channel: AbstractChannel,
    pipe: str,
    backoff_seconds: float = 0.1,
    max_retries: Optional[int] = None,
    clock: Clock = UtcClock(),
    getter: QueueGetterT = get_queue,
) -> SuccessT:
    queue = get_queue_name(pipe)
    if max_retries is None:
        return await _wait_for_queue_infinitely(
            channel, queue, backoff_seconds, clock, getter
        )
    elif max_retries == 0:
        return await _check_for_queue_once(channel, queue, getter)
    else:
        return await _wait_for_queue_with_backoff(
            channel, queue, max_retries, backoff_seconds, clock, getter
        )


def _calculate_backoff(
    retries: int,
    backoff_seconds: float,
    exponential_backoff: bool,
    max_backoff_seconds: Optional[float],
    apply_jitter: bool,
) -> float:
    wait_time_seconds = (
        backoff_seconds
        if not exponential_backoff
        else backoff_seconds * pow(2, retries)
    )
    if max_backoff_seconds is not None and wait_time_seconds > max_backoff_seconds:
        wait_time_seconds = max_backoff_seconds
    if apply_jitter:
        wait_time_seconds = random.uniform(0, wait_time_seconds)
    return wait_time_seconds


async def _wait_for_queue_infinitely(
    channel: AbstractChannel,
    queue: str,
    backoff_seconds: float,
    clock: Clock,
    getter: QueueGetterT,
) -> SuccessT:
    while True:
        try:
            await getter(channel, queue)
            return True
        except ChannelNotFoundEntity:
            await clock.sleep(backoff_seconds)


async def _check_for_queue_once(
    channel: AbstractChannel, queue: str, getter: QueueGetterT
) -> SuccessT:
    try:
        await getter(channel, queue)
    except ChannelNotFoundEntity:
        return False
    return True


async def _wait_for_queue_with_backoff(
    channel: AbstractChannel,
    queue: str,
    max_retries: int,
    backoff_seconds: float,
    clock: Clock,
    getter: QueueGetterT,
) -> SuccessT:
    success = False
    retries = 0
    while not success and retries < max_retries:
        try:
            await getter(channel, queue)
            success = True
        except ChannelNotFoundEntity:
            retries += 1
            await clock.sleep(backoff_seconds)
    return success



import asyncio
import threading
from typing import AsyncGenerator, Optional, cast, Any

import pytest
from aio_pika.abc import AbstractRobustConnection

from khnm.exceptions import NodeKwargsInvalid, PipelineDefinitionInvalid
from khnm.pipelines import make_pipeline
from tests.doubles import (
    generate_random_numbers_async,
    async_callback_stub,
    AsyncCallbackSpy,
    SampleDataObject,
)
from tests.utils import timeout


async def test_pipeline_sends_messages_source_to_sink(
    amqp_connection: AbstractRobustConnection,
) -> None:
    spy = AsyncCallbackSpy()

    pipeline = (
        make_pipeline()
        .add("test-source", generate_random_numbers_async)
        .add("test-node", async_callback_stub)
        .add("test-sink", spy)
        .build()
    )

    tasks = [
        asyncio.create_task(pipeline.run(amqp_connection, "test-source")),
        asyncio.create_task(pipeline.run(amqp_connection, "test-node")),
        asyncio.create_task(pipeline.run(amqp_connection, "test-sink")),
    ]

    async def spy_got_message() -> bool:
        return spy.received_obj is not None

    success = await timeout(spy_got_message, awaited_result=True, timeout_seconds=10)

    for task in tasks:
        task.cancel()

    assert success is True


async def test_pipeline_with_multiple_intermediate_nodes(
    amqp_connection: AbstractRobustConnection,
) -> None:
    spy = AsyncCallbackSpy()

    pipeline = (
        make_pipeline()
        .add("test-source", generate_random_numbers_async)
        .add("test-node-1", async_callback_stub)
        .add("test-node-2", async_callback_stub)
        .add("test-sink", spy)
        .build()
    )

    tasks = [
        asyncio.create_task(pipeline.run(amqp_connection, "test-source")),
        asyncio.create_task(pipeline.run(amqp_connection, "test-node-1")),
        asyncio.create_task(pipeline.run(amqp_connection, "test-node-2")),
        asyncio.create_task(pipeline.run(amqp_connection, "test-sink")),
    ]

    async def spy_got_message() -> bool:
        return spy.received_obj is not None

    success = await timeout(spy_got_message, awaited_result=True, timeout_seconds=10)

    for task in tasks:
        task.cancel()

    assert success is True


async def test_processing_through_pipeline_gives_correct_result(
    amqp_connection: AbstractRobustConnection,
) -> None:
    spy = AsyncCallbackSpy()

    async def generate_1() -> AsyncGenerator[SampleDataObject, None]:
        yield SampleDataObject(integer=2)

    def add(obj: SampleDataObject) -> SampleDataObject:
        return SampleDataObject(integer=obj.integer + 1)

    def multiply(obj: SampleDataObject) -> SampleDataObject:
        return SampleDataObject(integer=obj.integer * 3)

    pipeline = (
        make_pipeline()
        .add("number-source", generate_1)
        .add("add", add)
        .add("multiply", multiply)
        .add("result", spy)
        .build()
    )

    tasks = [
        asyncio.create_task(pipeline.run(amqp_connection, "number-source")),
        asyncio.create_task(pipeline.run(amqp_connection, "add")),
        asyncio.create_task(pipeline.run(amqp_connection, "multiply")),
        asyncio.create_task(pipeline.run(amqp_connection, "result")),
    ]

    async def get_result_from_spy() -> Optional[int]:
        obj = spy.received_obj
        if obj:
            result_ = cast(SampleDataObject, obj)
            return result_.integer
        return None

    result = await timeout(get_result_from_spy, awaited_result=9, timeout_seconds=10)

    for task in tasks:
        task.cancel()

    assert result == 9


async def test_processing_through_pipeline_gives_correct_result_with_custom_parameters(
    amqp_connection: AbstractRobustConnection,
) -> None:
    spy = AsyncCallbackSpy()

    async def generate_1() -> AsyncGenerator[SampleDataObject, None]:
        yield SampleDataObject(integer=2)

    def add(obj: SampleDataObject) -> SampleDataObject:
        return SampleDataObject(integer=obj.integer + 1)

    def multiply(obj: SampleDataObject) -> SampleDataObject:
        return SampleDataObject(integer=obj.integer * 3)

    pipeline = (
        make_pipeline()
        .add("number-source", generate_1, pipe_length=128)
        .add("add", add, pipe_length=256)
        .add("multiply", multiply, pipe_length=512)
        .add("result", spy)
        .build()
    )

    tasks = [
        asyncio.create_task(pipeline.run(amqp_connection, "number-source")),
        asyncio.create_task(pipeline.run(amqp_connection, "add")),
        asyncio.create_task(pipeline.run(amqp_connection, "multiply")),
        asyncio.create_task(pipeline.run(amqp_connection, "result")),
    ]

    async def get_result_from_spy() -> Optional[int]:
        obj = spy.received_obj
        if obj:
            result_ = cast(SampleDataObject, obj)
            return result_.integer
        return None

    result = await timeout(get_result_from_spy, awaited_result=9, timeout_seconds=10)

    for task in tasks:
        task.cancel()

    assert result == 9


async def test_node_runs_on_multiple_threads(
    amqp_connection: AbstractRobustConnection,
) -> None:
    threads: int = 4
    timeout_seconds: int = 5
    barrier = threading.Barrier(threads)
    observed_thread_ids = []
    lock = threading.Lock()
    spy = AsyncCallbackSpy()

    async def did_spy_got_result() -> bool:
        obj = spy.received_obj
        return obj is not None

    def slow_callback(obj: SampleDataObject) -> SampleDataObject:
        with lock:
            observed_thread_ids.append(threading.current_thread().ident)
        barrier.wait(timeout=timeout_seconds)
        return obj

    pipeline = (
        make_pipeline()
        .add("source", generate_random_numbers_async)
        .add("node", slow_callback)
        .add("sink", spy)
        .build()
    )

    tasks = [
        asyncio.create_task(pipeline.run(amqp_connection, "source")),
        asyncio.create_task(pipeline.run(amqp_connection, "node", threads=threads)),
        asyncio.create_task(pipeline.run(amqp_connection, "sink")),
    ]

    success = await timeout(
        did_spy_got_result,
        awaited_result=True,
        timeout_seconds=timeout_seconds * 2,
    )

    for task in tasks:
        task.cancel()

    assert success is True and len(set(observed_thread_ids)) == threads


async def test_sequential_node_is_detected_by_barrier(
    amqp_connection: AbstractRobustConnection,
) -> None:
    threads: int = 1
    timeout_seconds: int = 2
    barrier = threading.Barrier(threads + 1)  # always unsatisfiable with threads=1
    spy = AsyncCallbackSpy()

    async def did_spy_get_result() -> bool:
        return spy.received_obj is not None

    def slow_callback(obj: SampleDataObject) -> SampleDataObject:
        barrier.wait(timeout=timeout_seconds)
        return obj

    pipeline = (
        make_pipeline()
        .add("source", generate_random_numbers_async)
        .add("node", slow_callback)
        .add("sink", spy)
        .build()
    )

    tasks = [
        asyncio.create_task(pipeline.run(amqp_connection, "source")),
        asyncio.create_task(pipeline.run(amqp_connection, "node", threads=threads)),
        asyncio.create_task(pipeline.run(amqp_connection, "sink")),
    ]

    success = await timeout(
        did_spy_get_result,
        awaited_result=True,
        timeout_seconds=timeout_seconds * 2,
    )

    for task in tasks:
        task.cancel()

    assert success is not True
    assert barrier.broken


async def test_sink_runs_on_multiple_threads(
    amqp_connection: AbstractRobustConnection,
) -> None:
    threads: int = 4
    timeout_seconds: int = 5
    barrier = threading.Barrier(threads)
    observed_thread_ids = []
    lock = threading.Lock()
    processed = []

    def sink_callback(obj: SampleDataObject) -> None:
        with lock:
            observed_thread_ids.append(threading.current_thread().ident)
        barrier.wait(timeout=timeout_seconds)
        processed.append(True)

    pipeline = (
        make_pipeline()
        .add("source", generate_random_numbers_async)
        .add("sink", sink_callback)
        .build()
    )

    tasks = [
        asyncio.create_task(pipeline.run(amqp_connection, "source")),
        asyncio.create_task(pipeline.run(amqp_connection, "sink", threads=threads)),
    ]

    async def did_process() -> bool:
        return len(processed) > 0

    success = await timeout(
        did_process,
        awaited_result=True,
        timeout_seconds=timeout_seconds * 2,
    )

    for task in tasks:
        task.cancel()

    assert success is True and len(set(observed_thread_ids)) == threads


async def test_sequential_sink_is_detected_by_barrier(
    amqp_connection: AbstractRobustConnection,
) -> None:
    threads: int = 1
    timeout_seconds: int = 2
    barrier = threading.Barrier(threads + 1)  # always unsatisfiable with threads=1
    processed = []

    def sink_callback(obj: SampleDataObject) -> None:
        barrier.wait(timeout=timeout_seconds)
        processed.append(True)

    pipeline = (
        make_pipeline()
        .add("source", generate_random_numbers_async)
        .add("sink", sink_callback)
        .build()
    )

    tasks = [
        asyncio.create_task(pipeline.run(amqp_connection, "source")),
        asyncio.create_task(pipeline.run(amqp_connection, "sink", threads=threads)),
    ]

    async def did_process() -> bool:
        return len(processed) > 0

    success = await timeout(
        did_process,
        awaited_result=True,
        timeout_seconds=timeout_seconds * 2,
    )

    for task in tasks:
        task.cancel()

    assert success is not True
    assert barrier.broken


@pytest.mark.parametrize(
    "kwarg_name,kwarg_value",
    [
        ("connection_max_retries", 1),
        ("connection_backoff_seconds", 1),
        ("prefetch_count", 1),
    ],
)
async def test_source_node_kwargs_raises_exception_at_invalid_kwargs(
    amqp_connection: AbstractRobustConnection,
    kwarg_name: str,
    kwarg_value: Any,
) -> None:
    with pytest.raises(NodeKwargsInvalid):
        (
            make_pipeline()
            .add("source", generate_random_numbers_async, **{kwarg_name: kwarg_value})
            .add("sink", lambda obj: None)
            .build()
        )


@pytest.mark.parametrize(
    "kwarg_name,kwarg_value",
    [
        ("pipe_length", 1),
        ("durable", True),
        ("backoff_seconds", 1),
        ("max_retries", 1),
        ("exponential_backoff", True),
        ("max_backoff_seconds", 1.0),
        ("apply_jitter", True),
    ],
)
async def test_sink_node_kwargs_raises_exception_at_invalid_kwargs(
    amqp_connection: AbstractRobustConnection,
    kwarg_name: str,
    kwarg_value: Any,
) -> None:
    with pytest.raises(NodeKwargsInvalid):
        (
            make_pipeline()
            .add("source", generate_random_numbers_async)
            .add("sink", lambda obj: None, **{kwarg_name: kwarg_value})
            .build()
        )


async def test_pipeline_raises_exception_when_below_two_nodes(
    amqp_connection: AbstractRobustConnection,
) -> None:
    with pytest.raises(PipelineDefinitionInvalid):
        (make_pipeline().add("source", generate_random_numbers_async).build())



from aio_pika import Message
from aio_pika.abc import AbstractRobustConnection

from khnm.consumers import consume
from khnm.pipes import declare_pipe, send_message, get_queue_name
from tests.utils import get_queue_messages_count


async def test_consuming_message_removes_it_from_queue(
    amqp_connection: AbstractRobustConnection,
    sample_message: Message,
    pipe: str = "test",
) -> None:
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe)
        await send_message(channel, sample_message, pipe)
    async for message in consume(amqp_connection, pipe):
        async with message:
            break
    messages_count = await get_queue_messages_count(
        amqp_connection, get_queue_name(pipe)
    )
    assert messages_count == 0


async def test_on_exception_raised_the_message_is_requeued(
    amqp_connection: AbstractRobustConnection,
    sample_message: Message,
    pipe: str = "test",
) -> None:
    async with amqp_connection.channel() as channel:
        await declare_pipe(channel, pipe)
        await send_message(channel, sample_message, pipe)
    async for message in consume(amqp_connection, pipe):
        try:
            async with message:
                raise Exception()
        except Exception:
            break
    messages_count = await get_queue_messages_count(
        amqp_connection, get_queue_name(pipe)
    )
    assert messages_count == 1



from collections.abc import Awaitable, Callable
from contextlib import AbstractAsyncContextManager
from typing import Union, Iterable, TypeVar, AsyncGenerator, Generator, Protocol

import pydantic
from aio_pika import Message
from aio_pika.abc import AbstractChannel, AbstractQueue


type SuccessT = bool
type SenderT = Callable[[AbstractChannel, Message, str, bool], Awaitable[SuccessT]]
type QueueGetterT = Callable[[AbstractChannel, str], Awaitable[AbstractQueue]]
MessageObjectT = TypeVar("MessageObjectT", bound=pydantic.BaseModel)
type CallbackInputT = pydantic.BaseModel
type CallbackOutputT = Union[pydantic.BaseModel, Iterable[pydantic.BaseModel], None]
type SyncGeneratorCallbackT = Callable[..., Generator[CallbackOutputT, None, None]]
type AsyncGeneratorCallbackT = Callable[..., AsyncGenerator[CallbackOutputT]]
type GeneratorCallbackT = Union[SyncGeneratorCallbackT, AsyncGeneratorCallbackT]
type SyncCallbackT = Callable[..., CallbackOutputT]
type AsyncCallbackT = Callable[..., Awaitable[CallbackOutputT]]
type CallbackT = Union[
    SyncCallbackT,
    AsyncCallbackT,
]



import asyncio
import inspect
from asyncio import AbstractEventLoop
from collections.abc import Iterable
from concurrent.futures.thread import ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import (
    Self,
    Callable,
    Protocol,
    List,
    Union,
    AsyncGenerator,
    cast,
    Generator,
    Awaitable,
    Optional,
    TypedDict,
    Unpack,
    AsyncContextManager,
    get_type_hints,
)

import pydantic
from aio_pika.abc import AbstractRobustConnection, AbstractIncomingMessage
from pydantic import BaseModel

from khnm.consumers import consume
from khnm.exceptions import NodeKwargsInvalid, PipelineDefinitionInvalid
from khnm.producers import make_producer, Producer
from khnm.serialization import pydantic_model_to_message, message_to_pydantic_model
from khnm.time import Clock, UtcClock
from khnm.types import GeneratorCallbackT, CallbackT, CallbackOutputT


class Bag(pydantic.BaseModel):
    model_config = {"extra": "allow"}


class Runner(Protocol):
    @property
    def name(self) -> str:
        pass

    async def run(self, connection: AbstractRobustConnection, threads: int) -> None:
        pass


class Source(Runner):
    def __init__(
        self,
        name: str,
        downstream_pipe: str,
        callback: GeneratorCallbackT,
        pipe_length: Optional[int] = None,
        durable: bool = False,
        backoff_seconds: float = 0.1,
        max_retries: Optional[int] = None,
        exponential_backoff: bool = False,
        max_backoff_seconds: Optional[float] = None,
        apply_jitter: bool = False,
        clock: Clock = UtcClock(),
    ) -> None:
        self._name = name
        self._downstream_pipe = downstream_pipe
        self._callback = callback
        self._pipe_length = pipe_length
        self._durable = durable
        self._backoff_seconds = backoff_seconds
        self._max_retries = max_retries
        self._exponential_backoff = exponential_backoff
        self._max_backoff_seconds = max_backoff_seconds
        self._apply_jitter = apply_jitter
        self._clock = clock

    @property
    def name(self) -> str:
        return self._name

    async def run(self, connection: AbstractRobustConnection, threads: int) -> None:
        is_callback_async = _is_callback_async(self._callback)
        if is_callback_async:
            await self._run_async(connection)
        else:
            await self._run_sync(connection, threads)

    async def _run_async(self, connection: AbstractRobustConnection) -> None:
        await self._run_async_callback(connection)

    async def _run_sync(
        self, connection: AbstractRobustConnection, threads: int
    ) -> None:
        loop = asyncio.get_running_loop()
        with ThreadPoolExecutor(max_workers=threads) as executor:
            await self._run_sync_callback(connection, loop, executor)

    async def _run_async_callback(self, connection: AbstractRobustConnection) -> None:
        async with make_producer(
            connection,
            self._downstream_pipe,
            size=self._pipe_length,
            durable=self._durable,
            backoff_seconds=self._backoff_seconds,
            max_retries=self._max_retries,
            exponential_backoff=self._exponential_backoff,
            max_backoff_seconds=self._max_backoff_seconds,
            apply_jitter=self._apply_jitter,
            persistent_messages=self._durable,
            clock=self._clock,
        ) as producer:
            async for obj in cast(
                AsyncGenerator[CallbackOutputT, None], self._callback()
            ):
                if not isinstance(obj, pydantic.BaseModel):
                    raise TypeError(
                        f"Source callback must yield BaseModel instances, got {type(obj).__name__}"
                    )
                message = pydantic_model_to_message(obj)
                await producer.publish(message)

    async def _run_sync_callback(
        self,
        connection: AbstractRobustConnection,
        loop: AbstractEventLoop,
        executor: ThreadPoolExecutor,
    ) -> None:
        async with make_producer(
            connection,
            self._downstream_pipe,
            size=self._pipe_length,
            durable=self._durable,
            backoff_seconds=self._backoff_seconds,
            max_retries=self._max_retries,
            exponential_backoff=self._exponential_backoff,
            max_backoff_seconds=self._max_backoff_seconds,
            apply_jitter=self._apply_jitter,
            persistent_messages=self._durable,
            clock=self._clock,
        ) as producer:
            generator = cast(Generator[CallbackOutputT, None, None], self._callback())
            sentinel = object()
            while True:
                obj = await loop.run_in_executor(
                    executor, lambda: next(generator, sentinel)
                )
                if obj is sentinel:
                    break
                if not isinstance(obj, pydantic.BaseModel):
                    raise TypeError(
                        f"Source callback must yield BaseModel instances, got {type(obj).__name__}"
                    )
                message = pydantic_model_to_message(obj)
                await producer.publish(message)


class Node(Runner):
    def __init__(
        self,
        name: str,
        upstream_pipe: str,
        downstream_pipe: str,
        callback: CallbackT,
        pipe_length: Optional[int] = None,
        durable: bool = False,
        backoff_seconds: float = 0.1,
        max_retries: Optional[int] = None,
        exponential_backoff: bool = False,
        max_backoff_seconds: Optional[float] = None,
        apply_jitter: bool = False,
        connection_max_retries: Optional[int] = None,
        connection_backoff_seconds: float = 1.0,
        prefetch_count: Optional[int] = None,
        clock: Clock = UtcClock(),
    ) -> None:
        self._name = name
        self._upstream_pipe = upstream_pipe
        self._downstream_pipe = downstream_pipe
        self._callback = callback
        self._pipe_length = pipe_length
        self._durable = durable
        self._backoff_seconds = backoff_seconds
        self._max_retries = max_retries
        self._exponential_backoff = exponential_backoff
        self._max_backoff_seconds = max_backoff_seconds
        self._apply_jitter = apply_jitter
        self._connection_max_retries = connection_max_retries
        self._connection_backoff_seconds = connection_backoff_seconds
        self._prefetch_count = prefetch_count
        self._clock = clock

    @property
    def name(self) -> str:
        return self._name

    async def run(self, connection: AbstractRobustConnection, threads: int) -> None:
        is_callback_async = _is_callback_async(self._callback)
        if is_callback_async:
            await self._run_async(connection)
        else:
            await self._run_sync(connection, threads)

    async def _run_async(self, connection: AbstractRobustConnection) -> None:
        while True:
            await self._run_async_callback(connection)

    async def _run_sync(
        self, connection: AbstractRobustConnection, threads: int
    ) -> None:
        loop = asyncio.get_running_loop()
        with ThreadPoolExecutor(max_workers=threads) as executor:
            while True:
                await self._run_sync_callback(connection, loop, executor, threads)

    async def _run_async_callback(self, connection: AbstractRobustConnection) -> None:
        async with make_producer(
            connection,
            self._downstream_pipe,
            size=self._pipe_length,
            durable=self._durable,
            backoff_seconds=self._backoff_seconds,
            max_retries=self._max_retries,
            exponential_backoff=self._exponential_backoff,
            max_backoff_seconds=self._max_backoff_seconds,
            apply_jitter=self._apply_jitter,
            persistent_messages=self._durable,
            clock=self._clock,
        ) as producer:
            async for message in consume(
                connection,
                self._upstream_pipe,
                upstream_connection_max_retries=self._connection_max_retries,
                upstream_connection_backoff_seconds=self._connection_backoff_seconds,
                prefetch_count=self._prefetch_count,
                clock=self._clock,
            ):
                async with message as current_message:
                    obj = message_to_pydantic_model(current_message, Bag)
                    result = await cast(Awaitable[CallbackOutputT], self._callback(obj))
                    if result is not None:
                        await _handle_callback_output(result, producer)

    async def _run_sync_callback(
        self,
        connection: AbstractRobustConnection,
        loop: AbstractEventLoop,
        executor: ThreadPoolExecutor,
        threads: int,
    ) -> None:
        async with make_producer(
            connection,
            self._downstream_pipe,
            size=self._pipe_length,
            durable=self._durable,
            backoff_seconds=self._backoff_seconds,
            max_retries=self._max_retries,
            exponential_backoff=self._exponential_backoff,
            max_backoff_seconds=self._max_backoff_seconds,
            apply_jitter=self._apply_jitter,
            persistent_messages=self._durable,
            clock=self._clock,
        ) as producer:
            semaphore = asyncio.Semaphore(threads)
            pending = set()

            async def process(
                message_context: AsyncContextManager[AbstractIncomingMessage],
            ) -> None:
                async with semaphore:
                    async with message_context as current_message:
                        obj = message_to_pydantic_model(current_message, Bag)
                        result = cast(
                            CallbackOutputT,
                            await loop.run_in_executor(executor, self._callback, obj),
                        )
                        if result is not None:
                            await _handle_callback_output(result, producer)

            async for message in consume(
                connection,
                self._upstream_pipe,
                upstream_connection_max_retries=self._connection_max_retries,
                upstream_connection_backoff_seconds=self._connection_backoff_seconds,
                prefetch_count=self._prefetch_count,
                clock=self._clock,
            ):
                task = asyncio.create_task(process(message))
                pending.add(task)
                task.add_done_callback(pending.discard)

            if pending:
                await asyncio.gather(*pending)


class Sink(Runner):
    def __init__(
        self,
        name: str,
        upstream_pipe: str,
        callback: CallbackT,
        connection_max_retries: Optional[int] = None,
        connection_backoff_seconds: float = 1.0,
        prefetch_count: Optional[int] = None,
        clock: Clock = UtcClock(),
    ) -> None:
        self._name = name
        self._upstream_pipe = upstream_pipe
        self._callback = callback
        self._connection_max_retries = connection_max_retries
        self._connection_backoff_seconds = connection_backoff_seconds
        self._prefetch_count = prefetch_count
        self._clock = clock

    @property
    def name(self) -> str:
        return self._name

    async def run(self, connection: AbstractRobustConnection, threads: int) -> None:
        is_callback_async = _is_callback_async(self._callback)
        if is_callback_async:
            await self._run_async(connection)
        else:
            await self._run_sync(connection, threads)

    async def _run_async(self, connection: AbstractRobustConnection) -> None:
        while True:
            await self._run_async_callback(connection)

    async def _run_sync(
        self, connection: AbstractRobustConnection, threads: int
    ) -> None:
        loop = asyncio.get_running_loop()
        with ThreadPoolExecutor(max_workers=threads) as executor:
            while True:
                await self._run_sync_callback(connection, loop, executor, threads)

    async def _run_async_callback(self, connection: AbstractRobustConnection) -> None:
        async for message in consume(
            connection,
            self._upstream_pipe,
            upstream_connection_max_retries=self._connection_max_retries,
            upstream_connection_backoff_seconds=self._connection_backoff_seconds,
            prefetch_count=self._prefetch_count,
            clock=self._clock,
        ):
            async with message as current_message:
                obj = message_to_pydantic_model(current_message, Bag)
                await cast(Awaitable[CallbackOutputT], self._callback(obj))

    async def _run_sync_callback(
        self,
        connection: AbstractRobustConnection,
        loop: AbstractEventLoop,
        executor: ThreadPoolExecutor,
        threads: int,
    ) -> None:
        semaphore = asyncio.Semaphore(threads)
        pending = set()

        async def process(
            message_context: AsyncContextManager[AbstractIncomingMessage],
        ) -> None:
            async with semaphore:
                async with message_context as current_message:
                    obj = message_to_pydantic_model(current_message, Bag)
                    await loop.run_in_executor(executor, self._callback, obj)

        async for message in consume(
            connection,
            self._upstream_pipe,
            upstream_connection_max_retries=self._connection_max_retries,
            upstream_connection_backoff_seconds=self._connection_backoff_seconds,
            prefetch_count=self._prefetch_count,
            clock=self._clock,
        ):
            task = asyncio.create_task(process(message))
            pending.add(task)
            task.add_done_callback(pending.discard)

        if pending:
            await asyncio.gather(*pending)


class SinkKwargs(TypedDict, total=False):
    connection_max_retries: int
    connection_backoff_seconds: float
    prefetch_count: int


class SourceKwargs(TypedDict, total=False):
    pipe_length: int
    durable: bool
    backoff_seconds: float
    max_retries: int
    exponential_backoff: bool
    max_backoff_seconds: float
    apply_jitter: bool


class NodeKwargs(SourceKwargs, SinkKwargs):
    pass


RunnerKwargs = Union[SourceKwargs, SinkKwargs, NodeKwargs]


@dataclass(init=True, frozen=True)
class RunnerDefinition:
    name: str
    callback: Union[CallbackT, GeneratorCallbackT]
    kwargs: NodeKwargs = field(default_factory=NodeKwargs)


class PipelineBuilder:
    def __init__(self, name: Optional[str] = None) -> None:
        self._name = name
        self._runner_definitions: List[RunnerDefinition] = []

    def add(
        self,
        name: str,
        callback: Union[GeneratorCallbackT, CallbackT],
        **kwargs: Unpack[NodeKwargs],
    ) -> Self:
        if self._name:
            full_name = f"{self._name}.{name}"
        else:
            full_name = name
        self._runner_definitions.append(
            RunnerDefinition(
                name=full_name, callback=callback, kwargs=cast(NodeKwargs, kwargs)
            )
        )
        return self

    def build(self) -> Pipeline:
        if len(self._runner_definitions) < 2:
            raise PipelineDefinitionInvalid("Pipeline must have at least 2 nodes")
        definitions = list(self._runner_definitions)
        runners: List[Runner] = []
        source_definition = definitions.pop(0)
        next_node_definition = definitions.pop(0)
        _validate_source_kwargs(**source_definition.kwargs)
        runners.append(
            Source(
                name=source_definition.name,
                downstream_pipe=next_node_definition.name,
                callback=cast(GeneratorCallbackT, source_definition.callback),
                **cast(SourceKwargs, source_definition.kwargs),
            )
        )
        while definitions:
            current_node_definition = next_node_definition
            next_node_definition = definitions.pop(0)
            runners.append(
                Node(
                    name=current_node_definition.name,
                    upstream_pipe=current_node_definition.name,
                    downstream_pipe=next_node_definition.name,
                    callback=cast(CallbackT, current_node_definition.callback),
                    **current_node_definition.kwargs,
                )
            )
        _validate_sink_kwargs(**next_node_definition.kwargs)
        runners.append(
            Sink(
                name=next_node_definition.name,
                upstream_pipe=next_node_definition.name,
                callback=cast(CallbackT, next_node_definition.callback),
                **cast(SinkKwargs, next_node_definition.kwargs),
            )
        )
        return Pipeline(runners)


class Pipeline:
    def __init__(self, runners: List[Runner]) -> None:
        self._runners = runners
        self._runners_map = {runner.name: runner for runner in runners}

    async def run(
        self, connection: AbstractRobustConnection, name: str, threads: int = 1
    ) -> None:
        runner = self._runners_map.get(name)
        if not runner:
            raise ValueError(f"Runner {name} not found in the pipeline")
        await runner.run(connection, threads)


def make_pipeline(name: Optional[str] = None) -> PipelineBuilder:
    return PipelineBuilder(name)


async def _handle_callback_output(result: CallbackOutputT, producer: Producer) -> None:
    if _is_callback_result_iterable(result):
        results_serialized = [
            pydantic_model_to_message(item)
            for item in cast(Iterable[BaseModel], result)
        ]
        for item_serialized in results_serialized:
            await producer.publish(item_serialized)
    else:
        if not isinstance(result, BaseModel):
            raise TypeError(
                f"Callback must return a BaseModel instance, got {type(result).__name__}"
            )
        result_serialized = pydantic_model_to_message(result)
        await producer.publish(result_serialized)


def _is_callback_async(callback: Callable[..., ...]) -> bool:
    target = callback
    if not inspect.isfunction(callback) and not inspect.ismethod(callback):
        if hasattr(callback, "__call__"):
            target = callback.__call__
        else:
            return False

    return inspect.iscoroutinefunction(target) or inspect.isasyncgenfunction(target)


def _is_callback_result_iterable(
    value: CallbackOutputT,
) -> bool:
    return isinstance(value, Iterable) and not isinstance(value, BaseModel)


def _validate_source_kwargs(**kwargs: Unpack[NodeKwargs]) -> None:
    for kwarg_name, _ in kwargs.items():
        if kwarg_name not in get_type_hints(SourceKwargs):
            raise NodeKwargsInvalid(f"Invalid argument for source node: {kwarg_name}")


def _validate_sink_kwargs(**kwargs: Unpack[NodeKwargs]) -> None:
    for kwarg_name, _ in kwargs.items():
        if kwarg_name not in get_type_hints(SinkKwargs):
            raise NodeKwargsInvalid(f"Invalid argument for sink node: {kwarg_name}")



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager, AbstractAsyncContextManager
from typing import Optional, AsyncIterator

from aio_pika.abc import (
    AbstractRobustConnection,
    AbstractChannel,
    AbstractIncomingMessage,
)

from khnm.exceptions import UpstreamPipeUnavailable
from khnm.pipes import get_queue, wait_for_pipe, get_queue_name
from khnm.time import Clock, UtcClock
from khnm.types import QueueGetterT


async def consume(
    amqp_connection: AbstractRobustConnection,
    upstream_pipe: str,
    upstream_connection_max_retries: Optional[int] = None,
    upstream_connection_backoff_seconds: float = 1.0,
    upstream_queue_getter: QueueGetterT = get_queue,
    prefetch_count: Optional[int] = None,
    clock: Clock = UtcClock(),
) -> AsyncGenerator[AbstractAsyncContextManager[AbstractIncomingMessage], None]:
    channel = await _connect(
        connection=amqp_connection,
        upstream_pipe=upstream_pipe,
        upstream_connection_backoff_seconds=upstream_connection_backoff_seconds,
        upstream_connection_max_retries=upstream_connection_max_retries,
        prefetch_count=prefetch_count,
        queue_getter=upstream_queue_getter,
        clock=clock,
    )
    try:
        queue = await upstream_queue_getter(channel, get_queue_name(upstream_pipe))
        async with queue.iterator() as queue_iter:
            async for message in queue_iter:
                yield _handle_message(message)
    finally:
        await channel.close()


async def _connect(
    connection: AbstractRobustConnection,
    upstream_pipe: str,
    upstream_connection_backoff_seconds: float = 1.0,
    upstream_connection_max_retries: Optional[int] = None,
    prefetch_count: Optional[int] = None,
    queue_getter: QueueGetterT = get_queue,
    clock: Clock = UtcClock(),
) -> AbstractChannel:
    channel = await connection.channel()
    if prefetch_count is not None:
        await channel.set_qos(prefetch_count=prefetch_count)
    success = await wait_for_pipe(
        channel,
        upstream_pipe,
        upstream_connection_backoff_seconds,
        upstream_connection_max_retries,
        clock,
        queue_getter,
    )
    if not success:
        raise UpstreamPipeUnavailable(f"Upstream unavailable: {upstream_pipe}")
    return channel


@asynccontextmanager
async def _handle_message(
    message: AbstractIncomingMessage,
) -> AsyncIterator[AbstractIncomingMessage]:
    try:
        yield message
    except Exception:
        await message.nack(requeue=True)
        raise
    else:
        await message.ack()
